#ifndef THREAD_LOCAL_STORAGE_H
    #define THREAD_LOCAL_STORAGE_H

    #include <memory.h>

    #ifdef _WIN32
        #include <malloc.h>
        #include "ThreadLocalStorage_win32.h"
    #else
        #include "ThreadLocalStorage_posix.h"
    #endif

    class ThreadLocalStorageBase
    {
    public:

        static void Initialize()
        {
            //Must be called before any thread is started ( not thread safe at all )
            assert( !m_is_initialized );
            m_tls_identifier = TLSCreate();
            m_is_initialized = true;
        }

        static void Finalize()
        {
            //Must be called after all threads have been stopped ( not thread safe at all )
            TLSDestroy( m_tls_identifier );
        }

        static void StartupOnThreadEnter()
        {
            void * local_storage_table;

            local_storage_table = malloc( m_buffer_size );
            memset( local_storage_table, 0, m_buffer_size );

            TLSSet( m_tls_identifier, (void*)local_storage_table );

            ThreadLocalStorageBase * current = m_first;

            while( current )
            {
                current->InitializeObject();
                current = current->m_next;
            }
        }

        static void CleanupOnThreadExit()
        {
            ThreadLocalStorageBase * current = m_first;

            while( current )
            {
                current->FinalizeObject();
                current = current->m_next;
            }

            free( TLSGet( m_tls_identifier ) );
        }

    protected:

        enum{ Alignment = 4 };

        ThreadLocalStorageBase( int object_size )
        {
            assert( !m_is_initialized );
            object_size = ( object_size + Alignment - 1 ) & ~( Alignment - 1 );
            m_index = m_buffer_size;
            m_buffer_size += object_size;
            // Not Thread safe

            m_next = m_first;
            m_first = this;
        }

        void * GetThreadLocalStorage() const 
        {
            char * local_storage_table = reinterpret_cast<char*>( TLSGet( m_tls_identifier ) );

            return &local_storage_table[ m_index ];
        }

        virtual void InitializeObject() = 0;

        virtual void FinalizeObject() = 0;

        int m_index;
        ThreadLocalStorageBase * m_next;
        static int m_buffer_size;
        static TLSIndex m_tls_identifier;
        static ThreadLocalStorageBase * m_first;
        static bool m_is_initialized;
    };


    template< typename T >
    class ThreadLocalStorage : public ThreadLocalStorageBase
    {
    public:

        ThreadLocalStorage( const T & default_value = T() ) :
            ThreadLocalStorageBase( sizeof( T ) ),
            m_default_value( default_value )
        {

        }

        operator const T&() const
        {
            void * thread_local_storage = GetThreadLocalStorage();
            return *reinterpret_cast<T*>( thread_local_storage );
        }

        ThreadLocalStorage & operator=( const T & value )
        {
            void * thread_local_storage = GetThreadLocalStorage();
            *reinterpret_cast<T*>( thread_local_storage ) = value;
            return *this;
        }

    protected:

        virtual void InitializeObject()
        {
            new( GetThreadLocalStorage() ) T( m_default_value );
        }

        virtual void FinalizeObject()
        {
            static_cast<T*>( GetThreadLocalStorage() )->~T();
        }

        const T m_default_value;

    };

#endif
